package pathgame.tilemap;

import java.util.Random;
import java.util.TreeMap;

public class TileMapGenerator
{
    private static final TreeMap<Integer, Tile> GRASS_MAP = new TreeMap<>();
    
    static
    {
        GRASS_MAP.put(0, Tiles.GRASS);
        GRASS_MAP.put(76, Tiles.GRASS_WITH_STONE);
        GRASS_MAP.put(80, Tiles.GRASS_WITH_6_BUSHES);
        GRASS_MAP.put(84, Tiles.GRASS_WITH_3_BUSHES);
        GRASS_MAP.put(88, Tiles.GRASS_WITH_FLOWERS);
        GRASS_MAP.put(92, Tiles.GRASS_WITH_HILL);
        GRASS_MAP.put(96, Tiles.GRASS_WITH_EARTH);
    }
    
    private static long seed = 1;
    
    /** Returns a random generated map
     *
     * @param width the width of the map
     * @param height the height of the map
     * @return a random generated map
     */
    public static TileMap getMap(int width, int height)
    {
        return getMap(width, height, seed++);
    }
    
    /** Returns a random generated map
     *
     * @param width the width of the map
     * @param height the height of the map
     * @param seed the seed for the random number generator
     * @return a random generated map
     */
    public static TileMap getMap(int width, int height, long seed)
    {
        Random r = new Random(seed);
        
        HighMap highMap = HighMap.generate(seed, width, height);
        
        TileMap map = new TileMap(width, height);
        for(int x = 0; x < width; x++)
        {
            for(int y = 0; y < height; y++)
            {
                if(highMap.get(x, y) < 0.15f)
                {
                    map.setTile(x, y, Tiles.DEEP_WATER);
                }
                else if(highMap.get(x, y) < 0.3f)
                {
                    map.setTile(x, y, Tiles.SHALLOW_WATER);
                }
                else if(highMap.get(x, y) < 0.7f)
                {
                    map.setTile(x, y, randomGrass(r));
                }
                else if(highMap.get(x, y) < 0.85f)
                {
                    map.setTile(x, y, Tiles.HILL);
                }
                else
                {
                    map.setTile(x, y, Tiles.MOUNTAIN);
                }
            }
        }
        
        int forestSize = ((width + height) / 2) / 10;
        forestSize *= forestSize;
        
        generateForest(map, r, forestSize, 10, 2, Tiles.FOREST);
        generateForest(map, r, forestSize, 5, 2, Tiles.SWAMP);
        return map;
    }
    
    private static Tile randomGrass(Random r)
    {
        return GRASS_MAP.floorEntry(r.nextInt(100)).getValue();
    }
    
    private static void generateForest(TileMap map, Random r, int depth, int placements, int jumpRadius, Tile t)
    {
        for(int i = 0; i < placements; i++)
        {
            int x = r.nextInt(map.getWidth());
            int y = r.nextInt(map.getHeight());
            while(map.getTile(x, y).getForestReplaceChance() < 1.0f)
            {
                x = r.nextInt(map.getWidth());
                y = r.nextInt(map.getHeight());
            }
            
            for(int j = 0; j < depth; j++)
            {
                int oldX = x;
                int oldY = y;

                x += r.nextInt(jumpRadius * 2 + 1) - jumpRadius;
                y += r.nextInt(jumpRadius * 2 + 1) - jumpRadius;
                x = Math.min(Math.max(x, 0), map.getWidth() - 1);
                y = Math.min(Math.max(y, 0), map.getHeight() - 1);
                
                if(r.nextFloat() < map.getTile(x, y).getForestReplaceChance())
                {
                    map.setTile(x, y, t);
                    placeForest(map, r, x - 1, y, t);
                    placeForest(map, r, x + 1, y, t);
                    placeForest(map, r, x, y - 1, t);
                    placeForest(map, r, x, y + 1, t);
                }
                else
                {
                    x = oldX;
                    y = oldY;
                }
            }
        }
    }
    
    private static void placeForest(TileMap map, Random r, int x, int y, Tile t)
    {
        if(x >= 0 && x < map.getWidth() && y >= 0 && y < map.getHeight())
        {
            if(r.nextFloat() * 2 < map.getTile(x, y).getForestReplaceChance())
            {
                map.setTile(x, y, t);
            }
        }
    }
}
